# TRON 2.0 Migration Guide

`TRON 2.0` is a next major release, written in Swift 3. It is supported on following platforms:

* iOS 9.0
* macOS 10.11
* tvOS 9.0
* watchOS 2.0

If you are looking for Swift 2.2/Swift 2.3 compatible `TRON`, please use `TRON 1.2.x` - latest release, that is written in Swift 2.

### Alamofire 4

TRON 2.0 depends on Alamofire 4.0 release, which brought several big API changes to the table.

#### Data, Download and Upload requests

`APIRequest` and `MultipartAPIRequest` have been refactored to three distinct classes - `APIRequest`, `DownloadAPIRequest` and `UploadAPIRequest`, that directly map to Alamofire `DataRequest`, `DownloadRequest` and `UploadRequest`. As a result, `RequestType` enum was replaced by `UploadRequestType` and `DownloadRequestType` respectively.

#### Encoding

In TRON 1.0 there was a concept of encoding strategy - that determined which encoding to use based on HTTP Method. This concept is [built-in Alamofire itself](https://github.com/Alamofire/Alamofire#parameter-encoding), therefore `encodingStrategy` properties on `TRON` and `APIRequest` are replaced with simple `parameterEncoding` property that has `Alamofire.ParameterEncoding` type.

### Response serialization

`ResponseParseable` protocol was rewritten to directly inherit from `Alamofire.DataResponseSerializerProtocol`. It allows us to have two generic typealiases - `SerializedObject` from `Alamofire` and added to it `SerializedError` from TRON. It's really best to show them together: 

```swift

// Alamofire 4:

public protocol DataResponseSerializerProtocol {
    associatedtype SerializedObject

    var serializeResponse: (URLRequest?, HTTPURLResponse?, Data?, Error?) -> Result<SerializedObject> { get }
}

// TRON 2:

public protocol ErrorHandlingDataResponseSerializerProtocol : DataResponseSerializerProtocol {
    associatedtype SerializedError

    var serializeError: (Alamofire.Result<SerializedObject>?,URLRequest?, HTTPURLResponse?, Data?, Error?) -> APIError<SerializedError> { get }
}

```

This allows creating factory-like parsers, that support subclassed models, or use any parser that you want. It also, with [minor tweaking](https://github.com/MLSDev/TRON/blob/master/Tests/ResponseSerializationTestCase.swift#L15-L22), allows using **any** custom Alamofire response serializer with TRON directly!

`SwiftyJSONDecodable` protocol has been migrated to use new serialization protocol, so if you use it, there's no changes required.

`responseBuilder` and `errorBuilder` properties on `APIRequest` were replaced by `responseSerializer` and `errorSerializer` properties with new signatures.

### Better stubs

Previously, you would directly set success and error models on `APIStub`, which allowed to stub API, but testing your logic was not very convenient. In TRON 2, `APIStub` uses response and error serialization from `APIRequest`, that is containing that stub, thus allowing you to test response serialization.

### Improved plugins

In TRON 1.x plugin system was very limited, and had only two methods: 

```swift
// TRON 1.x

public protocol Plugin 
{
    func willSendRequest(_ request: URLRequest?)
    func requestDidReceiveResponse(_ response: (URLRequest?, HTTPURLResponse?, Data?, Error?))
}
```

In TRON 2.x, we are expanding `Plugin` protocol to have much more events, and are making them generic:

```swift
// TRON 2.x

public protocol Plugin {
    func willSendRequest<Model,ErrorModel>(_ request: BaseRequest<Model,ErrorModel>)

    func willSendAlamofireRequest<Model,ErrorModel>(_ request: Request, formedFrom tronRequest: BaseRequest<Model,ErrorModel>)

    func didSendAlamofireRequest<Model,ErrorModel>(_ request : Request, formedFrom tronRequest: BaseRequest<Model,ErrorModel>)

    func willProcessResponse<Model,ErrorModel>(response: (URLRequest?, HTTPURLResponse?, Data?, Error?),
                                                forRequest request: Request,
                                                formedFrom tronRequest: BaseRequest<Model,ErrorModel>)

    func didSuccessfullyParseResponse<Model,ErrorModel>(_ response: (URLRequest?, HTTPURLResponse?, Data?, Error?),
                                                        creating result: Model,
                                                        forRequest request: Request,
                                                        formedFrom tronRequest: BaseRequest<Model,ErrorModel>)

    func didReceiveError<Model,ErrorModel>(_ error: APIError<ErrorModel>,
                                        forResponse response : (URLRequest?, HTTPURLResponse?, Data?, Error?),
                                        request: Alamofire.Request,
                                        formedFrom tronRequest: BaseRequest<Model,ErrorModel>)

    func didReceiveDataResponse<Model,ErrorModel>(_ response: DataResponse<Model>, 
                                        forRequest request: Alamofire.Request, 
                                        formedFrom tronRequest: BaseRequest<Model,ErrorModel>)

    func didReceiveDownloadResponse<Model,ErrorModel>(_ response: DownloadResponse<Model>, 
                                                    forRequest request: Alamofire.DownloadRequest, 
                                                    formedFrom tronRequest: BaseRequest<Model,ErrorModel>)
}
```

This creates enormous space for plugin extensions. For example, you could build error handling system, that handles errors, that came from your API, both on global plugin level, and on local, if you need adjustments. You could see all your network requests lifespan by implementing something like this:

```swift
class NetworkMetricsPlugin : Plugin {
    func didReceiveDataResponse<Model, ErrorModel>(_ response: DataResponse<Model>, forRequest request: Request, formedFrom tronRequest: BaseRequest<Model, ErrorModel>) {
        debugPrint(request) // prints request cURL
        print(response.timeline) // prints request Timeline, collected by Alamofire
        print("---------------")
        print(response.metrics) // prints URLSessionTaskMetrics, new API, added in iOS 10 / macOS 10.12
    }
}
```

Note, however, that old `Plugin` methods are **no longer called**.

### Method naming

All API was adjusted to better match Swift 3 API design guidelines. Xcode built-in Swift Migrator should do almost all heavy-lifting, however you might need to review API changes manually in several cases.
